@@ -247,6 +247,20 @@ static int gok_sig_handler(int sig)
 	return -1;  /* this line gets rid of compiler warning */ 
 }
 
+/* callback for gok window state events */
+gboolean
+on_gok_window_state_event ( GtkWidget       *widget,
+							GdkEventWindowState  *event,
+							gpointer         user_data)
+{
+	if (event->changed_mask & GDK_WINDOW_STATE_ICONIFIED) {
+		if (event->new_window_state & GDK_WINDOW_STATE_ICONIFIED) {
+			gtk_window_present((GtkWindow*)widget);
+		}
+	}
+	return FALSE;
+}
+
 /**
 * main
 * @argc: ignored
@@ -315,6 +329,30 @@ gok_main_open(gint argc, gchar *argv[])
 		LIBGNOMEUI_PARAM_CRASH_DIALOG, FALSE,
 		NULL);
 
+	if (bLoginOption) {
+		GObject      *gobj;
+		BonoboObject *bobj;
+		int          ret;
+		const gchar  *display_name;
+		GSList       *reg_env = NULL;
+		
+		if (!bonobo_init (&argc, argv))
+			g_error ("Could not initialize oaf / Bonobo");
+
+		gobj = g_object_new (BONOBO_TYPE_APPLICATION, NULL);
+		bobj = BONOBO_OBJECT (gobj);
+
+		display_name = gdk_display_get_name (gdk_display_get_default ());
+		reg_env = bonobo_activation_registration_env_set (reg_env, "DISPLAY",
+								  display_name);
+
+		ret = bonobo_activation_register_active_server ("OAFIID:GNOME_Gok_Factory", bonobo_object_corba_objref (bobj), reg_env);
+		bonobo_activation_registration_env_free (reg_env);
+
+		if (ret != Bonobo_ACTIVATION_REG_SUCCESS)
+			return 1;
+	}
+	
 	/* Parse geometry */
 	if (geometry != NULL)
 	{
@@ -360,7 +398,8 @@ gok_main_open(gint argc, gchar *argv[])
 	/* initialize the word prediction engine with gconf values */
 	if (gok_data_get_use_aux_dictionaries ())
 		gok_wordcomplete_set_aux_dictionaries (gok_wordcomplete_get_default (),
-						       gok_data_get_aux_dictionaries ());
+						       (gchar *) gok_data_get_aux_dictionaries ());
+	/* cast silences compiler; we are discarding 'const' */
 
 	gok_log ("gok_data_initialize has finished");
 
@@ -368,6 +407,7 @@ gok_main_open(gint argc, gchar *argv[])
 	if (gok_data_get_use_gtkplus_theme () == FALSE)
 	{
 		gok_main_read_rc ();
+		gok_main_stop_theme_change_notify ();
 		gok_log ("gok_main_read_rc has finished");
 	}
 
@@ -441,7 +481,8 @@ gok_main_open(gint argc, gchar *argv[])
 	gok_chunker_initialize();
 	gok_keyboard_initialize();
 	gok_branchbackstack_initialize();
-	
+
+	fprintf (stderr, "login mode = %s\n", bLoginOption ? "true" : "false");
 	if (!bLoginOption) {
 		gok_main_initialize_wordcomplete ();
 		gok_main_initialize_commandprediction ();
@@ -495,7 +536,7 @@ gok_log ("check for currently active frame");
 gok_log ("finished check");
 	
 	/* connect signal handlers; don't trap SIGKILL */ 
-	signal(SIGSEGV, (void (*)(int))gok_sig_handler); 
+	signal(SIGSEGV, (void (*)(int))gok_sig_handler);
 	signal(SIGTERM, (void (*)(int))gok_sig_handler);
 	signal(SIGINT,  (void (*)(int))gok_sig_handler);
 
@@ -626,15 +667,43 @@ Accessible* gok_main_get_foreground_window_accessible()
 void gok_main_window_change_listener (Accessible* pAccessible)
 {
 	GokKeyboard* pKeyboard = NULL;
+	Accessible *acc = NULL;
 
 	gok_log_enter();	
 
 	/* note: ref counting happens in gok-spy's listeners */
 	m_pForegroundWindowAccessible = pAccessible;
+
+	/* reset the context menu until we receive a focus event */
+	/*gok_spy_set_context_menu_accessible (NULL, 0);*/
 	
-	/* we might currently have a dynamic keyboard so we need to branch to be safe */
-	gok_main_display_scan_previous_premade();
-	
+	if (pAccessible && Accessible_getChildCount (pAccessible) == 1
+	    && (acc = Accessible_getChildAtIndex (pAccessible, 0))
+	    && Accessible_getRole (acc) == SPI_ROLE_MENU) 
+	{
+	    AccessibleNode *node = g_malloc (sizeof (AccessibleNode));
+	    GokSpyUIFlags menu_flags;
+	    menu_flags.value = 0;
+	    menu_flags.data.menus = 1;
+	    node->paccessible = acc;
+	    node->flags.data.is_menu = 1;
+	    node->link = 0;
+	    node->pname = g_strdup (_("popup menu"));
+	    gok_spy_update_component_list (acc, menu_flags);
+	    gok_keyboard_branch_gui (node, GOK_SPY_SEARCH_MENU);
+	    g_free (node);
+	}
+	else
+	{
+	    /* we might currently have a dynamic keyboard so we need to branch to be safe */
+	    gok_main_display_scan_previous_premade();
+	}
+
+	if (acc) /* popup menu detection leaked this reference */
+	{
+	    Accessible_unref (acc); /* dont use gok_spy_accessible_unref since we didn't implicit ref above */
+	}
+
 	/*if (*/gok_keyboard_validate_dynamic_keys (pAccessible);/* == TRUE)*/
 	{
 		/* rechunk the keyboard because keys have changed state */
@@ -669,32 +738,54 @@ gok_main_object_state_listener (Accessible* pAccessible)
 	GokKeyboard* pKB;
 	GokKey* pK;
 	AccessibleStateSet* ass = NULL;
+	gboolean update_everything = FALSE;
 
 	pKB = gok_main_get_current_keyboard();
 	if (pKB != NULL) {
 		pK = pKB->pKeyFirst;
 		while (pK != NULL) {
-			if (pK->pNodeAccessible != NULL) {
-				if (pK->pNodeAccessible->paccessible == pAccessible) {
-					/* note: this is very specific for now because we are doing
-					   this to make sure checkbox and radio menu item state
-					   is represented properly. TODO: generalize
-					*/
-					ass = Accessible_getStateSet(pAccessible);
-					if (AccessibleStateSet_contains( ass, SPI_STATE_CHECKED )) {
-						pK->ComponentState.active = TRUE;
-					}
-					else {
-						pK->ComponentState.active = FALSE;
-					}
-					AccessibleStateSet_unref(ass);
-					gok_key_update_toggle_state (pK);
- 					break;
+			if (pK->accessible_node && (pK->accessible_node->paccessible == pAccessible)) {
+				/* note: this is very specific for now because we are doing
+				   this to make sure checkbox and radio menu item state
+				   is represented properly. TODO: generalize
+				*/
+				ass = Accessible_getStateSet(pAccessible);
+				if (AccessibleStateSet_contains( ass, SPI_STATE_CHECKED )) {
+					pK->ComponentState.active = TRUE;
 				}
+				else {
+					pK->ComponentState.active = FALSE;
+				}
+				if (!AccessibleStateSet_contains( ass, SPI_STATE_ENABLED ))
+				{
+					pK->Style = KEYSTYLE_INSENSITIVE;
+					gok_key_set_button_name (pK);
+				}
+				AccessibleStateSet_unref(ass);
+				gok_key_update_toggle_state (pK);
+				/* unfortunately, another special case for now... */
+				if (Accessible_getRole (pAccessible) == SPI_ROLE_PAGE_TAB) 
+				{
+					update_everything = TRUE;
+				}
+				break;
 			}
 			pK = pK->pKeyNext;
 		}
 	}
+
+	/* notes:
+	   we now check to see if the current keyboard requires updating, currently
+	   we simply call gok_main_ds, but if this is too slow we can create a 
+	   method that updates the specific key in question 
+	*/
+	if (update_everything) 
+	{
+		GokKeyboard *current_kbd = gok_main_get_current_keyboard ();
+		gok_spy_update_component_list (gok_main_get_foreground_window_accessible (), 
+					       current_kbd->flags);
+		gok_main_ds (current_kbd);
+	}
 }	
 
 
@@ -967,7 +1058,7 @@ gboolean gok_main_display_scan_previous_premade ()
 	while ((pKeyboard->bDynamicallyCreated == TRUE) && (index < MAX_DELETABLE_KEYBOARDS))
 	{
 		/* store the dynamic keyboards in the array */
-		gok_log("adding [%s] to the list of keyboards to delete", pKeyboard->Name);
+		gok_log("adding [%s] (%x) to the list of keyboards to delete", pKeyboard->Name, pKeyboard);
 		arrayKeyboardsToDelete[++index] = pKeyboard;
 		pKeyboard = gok_branchbackstack_pop();
 		if (pKeyboard == NULL)
@@ -982,10 +1073,10 @@ gboolean gok_main_display_scan_previous_premade ()
 		gok_log_leave();
 		return FALSE;
 	}
-		
+
 	/* display and scan the premade keyboard */
 	bReturnCode = gok_main_ds (pKeyboard);
-	
+
 	/* delete all the dynamic keyboards that were on the stack */
 	while (index >= 0)
 	{
@@ -1230,6 +1321,7 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 	return TRUE;
 }
 
+
 /**
 * gok_main_create_window
 * 
@@ -1296,6 +1388,10 @@ GtkWidget* gok_main_create_window (gboolean is_dock)
 	gtk_signal_connect (GTK_OBJECT (window1), "configure_event",
                       GTK_SIGNAL_FUNC (on_window1_configure_event),
                       NULL);
+					  
+	gtk_signal_connect (GTK_OBJECT (window1), "window_state_event",
+                      GTK_SIGNAL_FUNC (on_gok_window_state_event),
+                      NULL);
 
 	gok_log_leave();
 	return window1;
@@ -1338,6 +1434,37 @@ gok_main_read_rc ()
     }
 }
 
+/**
+ * gok_main_stop_theme_change_notify:
+ *
+ * stop the theme change monitoring by removing 
+ * the gtk-theme-name property from the GtkSetting object
+ * and reset the application to the default gtk theme
+ */
+
+gok_main_stop_theme_change_notify ()
+{
+  GdkDisplay *dpy = gdk_display_get_default ();
+  int i,j, num_scr = gdk_display_get_n_screens (dpy);
+
+  for (i=0; i < num_scr; i++)
+    {
+      GdkScreen *scr = gdk_display_get_screen (dpy, i);
+      GtkSettings *settings = gtk_settings_get_for_screen (scr);
+      GParamSpec* theme_name_spec = g_object_class_find_property (G_OBJECT_GET_CLASS (settings), "gtk-theme-name");
+      if (theme_name_spec)
+	{
+	  int ref_count;
+	  /* reset to default gtk theme */
+	  g_object_set (settings,"gtk-theme-name", "Default", NULL); 
+	  ref_count = theme_name_spec->ref_count;
+
+	  for (j = 0; j < ref_count; j++)
+	    g_param_spec_unref (theme_name_spec);
+	}
+    }
+}
+
 /**
 * gok_main_read_keyboards:
 *
@@ -1921,11 +2048,11 @@ gok_main_close()
 	gok_log("Stack pushes - pops:      [%d]",gok_branchbackstack_pushes_minus_pops());
 
 	gok_spy_close();
-	SPI_exit();
 
+	/* note, call all closing functions that delete UI before calling gok_settingsdialog_close */
+	gok_scanner_close(); /* deletes UI */
 	gok_settingsdialog_close();
 	
-	gok_scanner_close();
 	gok_data_close();
 
 	closeSwitchApi();
